"""
Functions to facilitate interactions with Optilogic platform using 'optilogic' library
"""

import os
import json
import logging
import time
from typing import Tuple
import optilogic


class OptilogicClient:
    """
    Wrapper for optilogic module for consumption in Cosmic Frog services
    """

    def __init__(self, username=None, appkey=None, logger=logging.getLogger()):
        # Detect if being run in Andromeda
        job_app_key = os.environ.get("OPTILOGIC_JOB_APPKEY")

        if appkey and not username:
            # Use supplied key
            self.api = optilogic.pioneer.Api(auth_legacy=False, appkey=appkey)
        elif appkey and username:
            # Use supplied key & name
            self.api = optilogic.pioneer.Api(auth_legacy=False, appkey=appkey, un=username)
        elif job_app_key:
            # Running on Andromeda
            self.api = optilogic.pioneer.Api(auth_legacy=False)
        else:
            raise ValueError("OptilogicClient could not authenticate")

        self.logger = logger

    def model_exists(self, model_name: str) -> bool:
        """
        Returns True if a given model exists, False otherwise
        """
        try:
            return self.api.storagename_database_exists(model_name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"Exception in cosmicfrog: {e}")
            return False

    def get_connection_string(self, model_name: str) -> Tuple[bool, str]:

        # TODO: There are two connection string fetch functions, see also frog_dbtools

        try:
            rv = {"message": "error getting connection string"}
            if not self.api.storagename_database_exists(model_name):
                return False, ""

            connection_info = self.api.sql_connection_info(model_name)

            return True, connection_info["connectionStrings"]["url"]

        except Exception as e:
            self.logger.error(f"Exception in cosmicfrog: {e}")
            return False, ""

    def create_model_synchronous(self, model_name: str, model_template: str):
        try:
            new_model = self.api.database_create(name=model_name, template=model_template)

            status = "success"
            rv = {}
            if "crash" in new_model:
                status = "error"
                rv["message"] = json.loads(new_model["response_body"])["message"]
                rv["httpStatus"] = new_model["resp"].status_code
            else:
                while not self.api.storagename_database_exists(model_name):
                    self.logger.info(f"creating {model_name}")
                    time.sleep(3.0)
                connections = self.api.sql_connection_info(model_name)
                rv["model"] = new_model
                rv["connection"] = connections

            return status, rv

        except Exception as e:
            return "exception", e



"""
    Functions for service authentication
"""

from typing import Dict
import os
import httpx
import jwt
import time
from .sync_wrapper import sync_wrapper
from .frog_log import get_logger


# Environment
SECURITY_TIMEOUT = os.getenv("SECURITY_TIMEOUT") or 20.0
ADMIN_APP_KEY = os.getenv("ADMIN_APP_KEY")  # Note: Deliberate, no defaults for urls
ATLAS_SERVICE_URL = os.getenv("ATLAS_SERVICE_URL")
ATLAS_API_BASE_URL = os.getenv("ATLAS_API_BASE_URL")
# the optilogic library has a setting for overriding the base
# url used for API calls. If this setting needs to be set, use
# this instead of the current ATLAS_API_BASE_URL
optilogic_api_url = os.getenv("OPTILOGIC_API_URL")
if optilogic_api_url:
    ATLAS_API_BASE_URL = f"{optilogic_api_url}/v0/"


def validate_jwt(jwt_token):
    """
    Offline validation for api token
    This will throw DecodeError if token cannot be decoded
    """
    jwt.decode(jwt_token, options={"verify_signature": False})


def validate_bearer_token(bearer_token):
    """
    Fast offline validation of bearer token
    Will throw if invalid, else can be checked online
    """
    assert bearer_token is not None, "Bearer token is missing"
    assert bearer_token.startswith("Bearer "), "Bearer token is missing prefix"
    assert bearer_token.strip() != "Bearer", "Bearer token is missing jwt"
    jwt_token = bearer_token[7:]
    validate_jwt(jwt_token)


def extract_credentials(*args):
    """
    Utility to support calling security related functions with a header or individual keys
    """
    if len(args) == 1 and hasattr(args[0], "get"):
        headers = args[0]
        app_key = headers.get("X-App-KEY", None)
        api_key = headers.get("X-API-KEY", None)
        bearer_token = headers.get("Authorization", None)
    else:
        app_key, api_key, bearer_token = args

    return app_key, api_key, bearer_token


def _make_header(app_key: str, api_key: str, bearer_token: str) -> Dict[str, str]:
    """
    Create header for platform api calls
    """

    if not (app_key or api_key or bearer_token):
        assert False, "No authentication was provided (all keys empty)"

    # Basic bearer key validation
    if not (app_key or api_key):
        validate_bearer_token(bearer_token)

    assert ATLAS_API_BASE_URL

    # set up header with app key or api key depending on value set
    if app_key:
        header_key = "X-APP-KEY"
    else:
        header_key = "X-API-KEY"

    return {header_key: app_key or api_key or bearer_token.replace("Bearer ", "")}


async def _get_app_key_async(*args):
    """
    Given an api key or bearer token, fetch an app key

    Warning: Ensure keys are authenticated before calling, will use sub claim
    """
    logger = get_logger()
    assert ADMIN_APP_KEY, "ADMIN_APP_KEY is not configured"
    assert ATLAS_SERVICE_URL, "ATLAS_SERVICE_URL is not configured"

    app_key, api_key, bearer_token = extract_credentials(*args)

    if app_key:
        return app_key

    if not api_key:
        validate_bearer_token(bearer_token)
        api_key = bearer_token.replace("Bearer ", "")
    else:
        validate_jwt(api_key)

    # Get client ID from the token
    decoded_data = jwt.decode(api_key, options={"verify_signature": False})

    data = {"userId": decoded_data["sub"], "name": "Issuing Cosmic Frog key"}

    headers = {"x-app-key": ADMIN_APP_KEY}
    rv = ""
    async with httpx.AsyncClient() as client:
        status_code = 0
        retry_count = 0
        s_max_retries = os.getenv("AUTH_MAX_RETRIES") or "3"
        max_retries = int(s_max_retries)

        while retry_count < max_retries and status_code != 200:
            retry_count +=1
            response = await client.post(
                f"{ATLAS_SERVICE_URL}", headers=headers, json=data, timeout=SECURITY_TIMEOUT
            )
            status_code = response.status_code

            if response.status_code == 409 and retry_count <= max_retries:
                logger.warning(f"retry {retry_count} to {ATLAS_SERVICE_URL} to get user token")
                time.sleep(5)
            elif response.status_code != 200:
                logger.error(f"Status code {response.status_code} for {ATLAS_SERVICE_URL}. Tried {retry_count} time(s)")
                raise ValueError(f"Status code {response.status_code} for {ATLAS_SERVICE_URL}. Tried {retry_count} time(s). Failed to get a token: {response.content}")
            else:
                rv = response.json().get("appkey")

    return rv


_get_app_key = sync_wrapper(_get_app_key_async)


async def _get_account_async(*args):
    """
    Fetch account details from platform

    Pass (app_key, api_key, bearer_token) or (request.headers)

    """

    assert ATLAS_API_BASE_URL, "ATLAS_API_BASE_URL is not configured"

    app_key, api_key, bearer_token = extract_credentials(*args)

    new_headers = _make_header(app_key, api_key, bearer_token)

    url = f'{ATLAS_API_BASE_URL.strip("/")}/account'

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=new_headers, timeout=SECURITY_TIMEOUT)
        response.raise_for_status()
        return response.json()


_get_account = sync_wrapper(_get_account_async)


async def socket_secured_async(app_key: str, api_key: str, bearer_token: str):
    """
    Used for socket.io
    """
    acc = await _get_account_async(app_key, api_key, bearer_token)

    assert acc

    return acc


socket_secured = sync_wrapper(socket_secured_async)



"""
Module with helper functions for working with sqlalchemy
"""

import os
import time
import sqlalchemy
from sqlalchemy.exc import OperationalError
from sqlalchemy import text

CFLIB_STATEMENT_TIMEOUT = (
    os.getenv("CFLIB_STATEMENT_TIMEOUT") or 1800000
)  # Statement timeout in milliseconds = 30 minutes
CFLIB_IDLE_TRANSACTION_TIMEOUT = os.getenv("CFLIB_IDLE_TRANSACTION_TIMEOUT") or 1800
CFLIB_CONNECT_TIMEOUT = os.getenv("CFLIB_CONNECT_TIMEOUT") or 15  # Connection timeout in seconds
CFLIB_DEFAULT_MAX_RETRIES = os.getenv("CFLIB_DEFAULT_MAX_RETRIES") or 5
CFLIB_DEFAULT_RETRY_DELAY = os.getenv("CFLIB_DEFAULT_RETRY_DELAY") or 5


def create_engine_with_retry(
    logger,
    connection_string: str,
    application_name: str,
    max_retries: int = CFLIB_DEFAULT_MAX_RETRIES,
    retry_delay: int = CFLIB_DEFAULT_RETRY_DELAY,
) -> sqlalchemy.Engine:
    """
    Wrapper for sqlalchemy.create_engine - adds ping with retries, to ensure connection is valid for use
    """

    connection_string = f"{connection_string}&application_name={application_name}"

    engine = sqlalchemy.create_engine(
        connection_string,
        connect_args={"connect_timeout": CFLIB_CONNECT_TIMEOUT},
        execution_options={"statement_timeout": CFLIB_STATEMENT_TIMEOUT},
    )

    for attempt in range(max_retries):
        try:
            # Attempt to establish a connection and successful ping
            with engine.connect() as connection:
                connection.execute(text("SELECT 1;"))
                return engine

        except OperationalError:
            logger.warning("create_engine_with_retry: Database not ready, retrying")
            if attempt < max_retries - 1:
                # TODO: Async
                time.sleep(retry_delay)
            else:
                raise  # Re-raise the exception if all retries fail



"""
    Helper class for utility parameters
"""

from typing import List, Dict

POSSIBLE_TYPES = [
    "int",  # User enters an integer
    "double",  # User enters a double
    "string",  # User enters a string
]

# Note: Type can also be list of strings: [Option1, Option2] which requires the user to select


class Params:
    """
    A helper class to represent a set of parameters for a CF Model Function.

    Attributes
    ----------
    None

    Methods
    -------
    add - Adds a new parameter
    result - Returns all added parameters as a string
    """

    def __init__(self, params: List | Dict = None):
        self.__params = []
        self.app_key = None

        # Note: No validation on params passed here, is responsibility of the running script to validate
        if params is not None:
            # Old utilities (list of values) - TO REMOVE WHEN UNUSED
            if isinstance(params, list):
                self.__loaded_params = params
                self.model_name = self.__loaded_params.pop(0)

            # New utilities (dict of values)
            if isinstance(params, dict):
                self.__loaded_params = params
                self.model_name = self.__loaded_params["model_name"]
                self.__loaded_params.remove("model_name")

    def __getitem__(self, index):
        return self.__loaded_params[index]

    def __is_valid_param_type(self, param_type: str) -> bool:
        if param_type in POSSIBLE_TYPES:
            return True

        if param_type.startswith("[") and param_type.endswith("]"):
            return True

        return False

    def add(self, name: str, description: str, default: any, param_type: str) -> None:
        """Add a new parameter.

        Args:
            name: the name of the parameter to add.
            description: a user friendly description of the parameter
            default: default value of the parameter
            param_type: the type of the parameter
        Returns:
            None.
        Raises:
            ValueError if the type does not validate as a type
        """

        name = name.strip()
        description = description.strip()
        param_type = param_type.strip()
        if not param_type.startswith("["):
            param_type = param_type.lower()

        if not self.__is_valid_param_type(param_type):
            raise ValueError(f"Invalid param type: {param_type}")

        self.__params.append(
            {
                "Name": name,
                "Description": description,
                "Value": default,
                "Type": param_type,
            }
        )

    def result(self):
        """Returns the parameter list as a string.

        Args:
            None.
        Returns:
            String representation of added parameters.
        """
        return self.__params



from .frog_data import FrogModel
from .frog_log import get_logger
from .frog_params import Params
from .frog_platform import OptilogicClient
from .frog_launcher import start_utility
from .frog_utilitydetails import UtilityDetails
from .frog_security import *
from .frog_dbtools import *
from .frog_utils import FrogUtils



"""
    Helper functions for working with Cosmic Frog models
"""

import os
import sys
import time
import uuid
from typing import Dict, List, Optional, Tuple
from contextlib import contextmanager
from io import StringIO
from collections.abc import Iterable
from pandas import DataFrame
import pandas as pd
import numpy as np
import sqlalchemy
from sqlalchemy import text, inspect
from psycopg2 import sql
from psycopg2.errors import LockNotAvailable, DeadlockDetected
from .frog_platform import OptilogicClient
from .frog_log import get_logger
from .frog_activity import ModelActivity, ActivityStatus, activity_signal
from .db_helpers import create_engine_with_retry
from .frog_utils import FrogUtils
from .frog_anura import AnuraSchema

# pylint: disable=logging-fstring-interpolation

# TODO:
# Will need extensions for custom tables in a model
# Profile parallel write for xlsx
# Add batching to standard table writing

# Define chunk size (number of rows to write per chunk)
CHUNK_SIZE = 1000000

# For key columns, replace Null with placeholder
# For matching on key columns only, will not be written to final table!
PLACEHOLDER_VALUE = ""  # replace with a value that does not appear in your data

CFLIB_IDLE_TRANSACTION_TIMEOUT = os.getenv("CFLIB_IDLE_TRANSACTION_TIMEOUT") or 1800


class FrogModel:
    """
    FrogModel class with helper functions for accessing Cosmic Frog models
    """

    # This allows app key to be set once for all instances, makes utilities easier to write
    class_app_key = None

    def __init__(
        self,
        model_name: Optional[str] = None,
        connection_string: Optional[str] = None,
        engine: Optional[sqlalchemy.engine.Engine] = None,
        application_name: str = "CosmicFrog User Library",
        app_key: str = None,
    ) -> None:
        self.model_name = model_name
        self.engine = None
        self.connection = None
        self.transactions = []
        self.log = get_logger()
        self.default_schema = None

        # App key can be supplied in 4 ways:
        # 1) Passed in argument when opening a model e.g. FrogModel(app_key="my_app_key")
        # 2) Set via class variable (used for all instances of FrogModel, used in utilities)
        # 3) Via Enviroment var, in Andromeda
        # 4) Via app.key file (when running locally, place file in folder with your script)

        # Store an app key if supplied (it may not be if initialising via engine or connection string)
        if app_key:
            found_app_key = app_key
        elif FrogModel.class_app_key:
            found_app_key = FrogModel.class_app_key
        else:
            found_app_key = os.environ.get("OPTILOGIC_JOB_APPKEY")

        if not found_app_key:
            initial_script_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
            file_path = os.path.join(initial_script_dir, "app.key")

            # If local file 'app.key' exists then assume it contains a valid app key
            if os.path.exists(file_path):
                with open(file_path, "r", encoding="utf-8") as file:
                    found_app_key = file.read().strip()

        self._app_key = found_app_key

        # Model connection can happen in 3 ways:
        # 1) A model name is supplied (required app key) - connection string will be fetched
        # 2) A pre-connected engine is supplied
        # 3) A model connection string is supplied

        if model_name and not (connection_string or engine):

            oc = OptilogicClient(appkey=self._app_key)

            success, connection_string = oc.get_connection_string(model_name)

            if not success:
                raise ValueError(f"Cannot get connection string for frog model: {model_name}")

            self.engine = create_engine_with_retry(self.log, connection_string, application_name)

        # Initialise connection
        elif engine:
            self.engine = engine
        elif connection_string:
            self.engine = create_engine_with_retry(self.log, connection_string, application_name)

        # Identify Anura version
        self.anura_version = self.get_anura_version()

    def __enter__(self):
        self.start_transaction()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            # No exceptions occurred, so commit the transaction
            self.commit_transaction()
        else:
            # An exception occurred, so roll back the transaction
            self.rollback_transaction()

    # Note: The following are for user managed transactions (do not use for library internal transactions)
    def start_transaction(self) -> None:
        """Start a new transaction."""
        if self.connection is None:
            self.connection = self.engine.connect()

        if self.transactions:
            self.transactions.append(self.connection.begin_nested())
        else:
            self.transactions.append(self.connection.begin())

    def commit_transaction(self) -> None:
        """Commit the outermost transaction."""
        if self.transactions:
            transaction = self.transactions.pop()
            transaction.commit()

            if not self.transactions:
                self.connection.close()
                self.connection = None

    def rollback_transaction(self) -> None:
        """Rollback the outermost transaction."""
        if self.transactions:
            transaction = self.transactions.pop()
            transaction.rollback()

            if not self.transactions:
                self.connection.close()
                self.connection = None

    @contextmanager
    def begin(self):
        """Context manager for model connections, used to manage both user and system transactions"""
        connection = None
        try:
            # Decide the context based on the transaction state
            if self.transactions:  # If user has opened a transaction, then nest one
                connection = self.transactions[-1].connection
                transaction = connection.begin_nested()
            else:  # else start a new one
                connection = self.engine.connect()
                transaction = connection.begin()

            # IMPORTANT DETAIL -> yields the connection, not the transaction
            yield connection

            transaction.commit()  # commit the transaction if everything goes well

        except LockNotAvailable:
            self.log.warning("Rolling back, unable to execute due to existing locks on model")
            transaction.rollback()

        except DeadlockDetected:
            self.log.warning("Rolling back, DEADLOCK was detected during operation on model")
            transaction.rollback()

        except Exception:
            self.log.exception(
                "Error, rolling back transaction due to exception",
                exc_info=True,
                stack_info=True,
            )
            transaction.rollback()
            raise
        finally:
            # If the connection was created in this method, close it.
            if not self.transactions:
                connection.close()

    def get_anura_version(self):
        """Return the current Anura schema version"""
        df = self.read_sql("SELECT current_schema()")
        return df.iloc[0, 0]

    def get_tablelist(
        self,
        input_only: bool = False,
        output_only: bool = False,
        technology_filter: str = None,
        original_names: bool = False,
    ) -> List[str]:
        """Get a list of commonly used Anura tables, with various filtering options.

        Args:
        input_only:         Return only input tables
        output_only:        Return only output tables
        technology_filter:  Return tables matching technology (e.g. "NEO")
        original_names:     Return original (UI) names (e.g. "CustomerDemand" rather than "customerdemand")
        """
        assert not (input_only and output_only), "input_only and output_only cannot both be True"

        if technology_filter:
            filtered_data = [
                field
                for field in AnuraSchema.get_anura_masterlist(self.anura_version)
                if (
                    (technology_filter.upper() in field["Technology"].upper())
                    and (
                        (input_only and not field["Category"].startswith("Output"))
                        or (output_only and field["Category"].startswith("Output"))
                        or (not input_only and not output_only)
                    )
                )
            ]

            return [
                field["Table"].lower() if not original_names else field["Table"]
                for field in filtered_data
            ]

        lower_case = not original_names

        # Common un filtered cases
        if input_only:
            return AnuraSchema.get_anura_input_table_names(self.anura_version, lower_case)

        if output_only:
            return AnuraSchema.get_anura_output_table_names(self.anura_version, lower_case)

        return AnuraSchema.get_anura_table_names(self.anura_version, lower_case)

    def get_columns(self, table_name: str) -> List[str]:
        """List Anura columns for the given table

        Args:
        table_name: The target table to fetch columns for
        """

        lower_name = table_name.lower()

        return AnuraSchema.get_anura_columns(self.anura_version, lower_name)

    def get_key_columns(self, table_name: str) -> List[str]:
        """List Anura 'key' columns for the given table

        Args:
        table_name: The target table to fetch keys for
        """

        lower_name = table_name.lower()

        return AnuraSchema.get_anura_keys(self.anura_version, lower_name)

    # Dump data to a model table
    def write_table(
        self, table_name: str, data: pd.DataFrame | Iterable, overwrite: bool = False
    ) -> None:
        """Pushes data into a model table from a data frame or iterable object

        Args:
        table_name: The target table
        data:       The data to be written
        overwrite:  Set to true to overwrite current table contents
        """

        table_name = table_name.lower().strip()

        self.log.info("write_table, writing to: %s", table_name)

        # TODO: Should be under same transaction as the write
        if overwrite:
            self.clear_table(table_name)

        if isinstance(data, pd.DataFrame) is False:
            data = pd.DataFrame(data)

        data.columns = data.columns.astype(str).str.lower().map(str.strip)

        # Initial implementation - pull everything into a df and dump with to_sql
        with self.begin() as connection:
            data.to_sql(
                table_name,
                con=connection,
                if_exists="append",
                index=False,
                method="multi",
                chunksize=CHUNK_SIZE,
            )

        # Note: tried a couple of ways to dump the generator rows directly, but didn't
        # give significant performance over dataframe (though may be better for memory)
        # Decided to leave as is for now

    def read_table(self, table_name: str, id_col: bool = False) -> DataFrame:
        """Read a single model table and return as a DataFrame

        Args:
            table_name: Table name to be read (supporting custom tables)
            id_col: Indicates whether the table id column should be returned

        Returns:
            Single dataframe holding table contents
        """

        table_name = table_name.lower().strip()

        with self.begin() as connection:
            result = pd.read_sql(table_name, con=connection)
            if not id_col:
                result.drop(columns=["id"], inplace=True)
            return result

    # Read all, or multiple Anura tables
    def read_tables(
        self, table_list: List[str] = None, id_col: bool = False
    ) -> Dict[str, pd.DataFrame]:
        """Read multiple Anura tables and return as a dictionary, indexed by table name

        Args:
            table_list: List of table names to be read (supporting custom tables)
            id_col: Indicates whether the table id column should be returned

        Returns:
            Dictionary of tables, where key is table name and value is dataframe of contents
        """

        result = {}

        for t in table_list:
            result[t] = self.read_table(t, id_col=id_col)

        return result

    def clear_table(self, table_name: str):
        """Clear table of all content

        Args:
            table_name: Name of table to be cleared

        Returns:
            None
        """

        table_name = table_name.lower().strip()

        # delete any existing data data from the table
        self.exec_sql(f"TRUNCATE {table_name}")

        return True

    def geocode_table(
        self,
        table_name: str,
        geoprovider: str = "MapBox",
        geoapikey: str = None,
        ignore_low_confidence: bool = True,
    ):
        """
        Geocode a geocodable entity table (Customers, Facilities, Suppliers) in model
        """

        # TODO: This may fail if FrogModel initialised with no app key - i.e. from connection string or engine

        return FrogUtils.geocode_table(
            self.model_name,
            table_name,
            self._app_key,
            geoprovider=geoprovider,
            geoapikey=geoapikey,
            ignore_low_confidence=ignore_low_confidence,
        )

    # Read from model using raw sql query
    def read_sql(self, query: str) -> DataFrame:
        """
        Executes a sql query on the model and returns the results in a dataframe

        Args:
            query: SQL query to be run

        Returns:
            Dataframe containing results of query
        """
        with self.begin() as connection:
            return pd.read_sql_query(query, connection)

    # Execute raw sql on model
    def exec_sql(self, query: str | sql.Composed) -> None:
        with self.begin() as connection:
            connection.execute(text(query))

    # Upsert from a csv file to a model table
    def upsert_csv(
        self,
        table_name: str,
        filename: str,
        _activity: ModelActivity = None,
        _correlation_id: str = "",
    ) -> (int, int):
        """
        Upsert a csv file to a Cosmic Frog model table

        Args:
            table_name: Name of the target Anura table
            filename: Name of csv file to be imported

        Returns:
            updated_rows, inserted_rows
        """
        total_updated = 0
        total_inserted = 0

        try:
            file_size = os.path.getsize(filename)

            if file_size <= 0:
                self.log.warning("CSV file has no rows")
                return 0, 0

            with open(filename, "rb") as file_handle:
                for chunk in pd.read_csv(
                    file_handle, chunksize=CHUNK_SIZE, dtype=str, skipinitialspace=True
                ):

                    # Get the current file position in bytes
                    current_position = file_handle.tell()

                    chunk.replace("", np.nan, inplace=True)
                    updated, inserted = self.upsert(table_name, chunk, _correlation_id=_correlation_id, activity=_activity)

                    total_updated += updated
                    total_inserted += inserted

                    if _activity:
                        # TODO: Support async here

                        progress_pct = (current_position / file_size) * 100

                        _activity.update_activity(
                            ActivityStatus.STARTED,
                            last_message=f"Uploading csv to {table_name}",
                            progress=int(progress_pct),
                        )

                    activity_signal(
                        self.log,
                        f"TABLE INSERT {table_name} {inserted}",
                        app_key=self._app_key,
                        model_name=self.model_name,
                        correlation_id=_correlation_id,
                    )

            return total_updated, total_inserted
        except Exception as e:
            self.log.exception("Error upserting csv to model", exc_info=True)
            if _activity.status != ActivityStatus.FAILED:
                _activity.update_activity(
                    ActivityStatus.FAILED,
                    last_message=f"File upsert failed",
                    progress=100,
                )

    # Upsert from an xls file to a model table
    def upsert_excel(
        self, filename: str, _activity: ModelActivity = None, _correlation_id: str = ""
    ) -> (int, int):
        """
        Upsert an xlsx file to a Cosmic Frog model table

        Args:
            table_name: Name of the target Anura table
            filename: Name of xlsx file to be imported

        Returns:
            updated_rows, inserted_rows
        """

        # TODO: If an issue could consider another way to load/stream from xlsx maybe?

        try:
            with pd.ExcelFile(filename) as xls:
                file_name_without_extension = (
                    os.path.basename(filename).replace(".xlsx", "").replace(".xls", "")
                )

                total_sheets = len(xls.sheet_names)

                if total_sheets == 0:
                    self.log.warning("Excel file has no sheets")
                    return 0, 0

                # For each sheet in the file
                for count, sheet_name in enumerate(xls.sheet_names):
                    if _activity:
                        progress_pct = (count / total_sheets) * 100

                        # TODO: Support async here
                        _activity.update_activity(
                            ActivityStatus.STARTED,
                            last_message=f"Uploading {sheet_name}",
                            progress=int(progress_pct),
                        )

                    table_to_upload = (
                        file_name_without_extension if sheet_name[:5].lower() == "sheet" else sheet_name
                    )

                    # Read the entire sheet into a DataFrame
                    # Note: For xlsx there is an upper limit of ~1million rows per sheet, so not chunking here

                    # TODO: Consider parallelism
                    df = pd.read_excel(xls, sheet_name=sheet_name, dtype=str)

                    df.columns = df.columns.str.lower().map(str.strip)

                    df.replace("", np.nan, inplace=True)

                    updated, inserted = self.upsert(
                        table_to_upload, df, _correlation_id=_correlation_id, activity=_activity
                    )

                    if _activity:
                        msg = f"TABLE INSERT {table_to_upload} {inserted}"
                        self.log.debug(f"Signalling: {msg}")

                        activity_signal(
                            self.log,
                            msg,
                            app_key=self._app_key,
                            model_name=self.model_name,
                            correlation_id=_correlation_id,
                        )

            return updated, inserted
        except Exception as e:
            self.log.exception("Error upserting xlsx to model", exc_info=True)
            if _activity.status != ActivityStatus.FAILED:
                _activity.update_activity(
                    ActivityStatus.FAILED,
                    last_message=f"File upsert failed",
                    progress=100,
                )

    def get_table_columns_from_model(self, table_name: str, id_col: bool = False) -> List[str]:
        """
        Fetches all columns direct from database (including custom)

        This gets all actual columns in the model table, including user custom columns
        """
        table_name = table_name.lower().strip()

        # Create an Inspector object
        inspector = inspect(self.engine)

        # Get the column names for a specific table
        column_names = inspector.get_columns(table_name)

        column_names = [column["name"] for column in column_names]

        if not id_col:
            column_names.remove("id")

        return [name.lower().strip() for name in column_names]

    def _get_combined_key_columns_for_upsert(self, table_name: str):

        anura_keys = AnuraSchema.get_anura_keys(self.anura_version, table_name)

        # TODO: Call platform custom columns apis here
        custom_key_columns = ["notes"]
        
        # look at platform api for custom columns in table for that model, if they exist then add them to custom_key_columns

        return anura_keys + custom_key_columns

    def __generate_index_sql(self, index_name, table_name, key_column_list) -> str:
        """
        Creates an appropriate index for Anura tables.
        Coalesce is used to support
        """
        coalesced_columns = ", ".join(
            [f"COALESCE({column}, '{PLACEHOLDER_VALUE}')" for column in key_column_list]
        )

        return f"CREATE INDEX {index_name} ON {table_name}({coalesced_columns});"

    def _create_upsert_index_for_table(
        self, table_name: str, cursor, combined_key_columns=None
    ) -> str:
        """
        Creates an index to aid update/insert performance for upsert
        """

        if combined_key_columns is None:
            combined_key_columns = self._get_combined_key_columns_for_upsert(table_name)

        upsert_index_name = "cf_upsert_index_" + str(uuid.uuid4()).replace("-", "")
        index_sql = self.__generate_index_sql(upsert_index_name, table_name, combined_key_columns)

        start_time = time.time()
        cursor.execute(index_sql)
        end_time = time.time()
        self.log.info(
            f"Index creation took {end_time - start_time} seconds for {table_name}",
        )

        return upsert_index_name

    def upsert(
        self,
        table_name: str,
        data: pd.DataFrame,
        _correlation_id: str = "",  # Optional: correlation id for logging / tracing (for internal use)
        activity: ModelActivity = None,  # Optional: activity object for progress updates
    ) -> Tuple[int, int]:
        """
        Upsert a pandas dataframe to a Cosmic Frog model table

        Args:
            table_name: Name of the target Anura table
            data: A Pandas dataframe containing the data to upsert

        Returns:
            updated_rows, inserted_rows
        """

        if len(data) <= 0:
            self.log.warning("Aborting upsert. Input dataframe is empty")
            return 0, 0

        table_name = table_name.strip().lower()

        data.columns = data.columns.str.lower().map(str.strip)

        anura_tables = AnuraSchema.get_anura_table_names(self.anura_version)

        table_exists = any(s.lower() == table_name for s in anura_tables)

        if not table_exists:
            # Skip it
            self.log.warning(
                "Table name not recognised as an Anura table (skipping): %s", table_name
            )
            return 0, 0

        self.log.info("Importing to table: %s", table_name)
        self.log.info("Source data has %s rows", len(data))

        # Behavior rules:
        # Key columns - get used to match (Note: possible future requirement, some custom columns may also be key columns)
        # Other Anura columns - get updated
        # Other Custom columns - get updated
        # Other columns (neither Anura or Custom) - get ignored

        all_column_names = self.get_table_columns_from_model(table_name)
        if "id" in all_column_names:
            all_column_names.remove("id")

        # 1) Anura key cols - defined in Anura
        # 2) Custom key cols - TBD CF Meta
        # 3) Update cols - The rest

        combined_key_columns = self._get_combined_key_columns_for_upsert(table_name)

        update_columns = [col for col in all_column_names if col not in combined_key_columns]

        # Skipping unrecognised columns (Do not trust column names from user data)
        cols_to_drop = [col for col in data.columns if col not in all_column_names]

        for col in cols_to_drop:
            self.log.info("Skipping unknown column in %s: %s", table_name, col)

        data = data.drop(cols_to_drop, axis=1)

        before_rows = len(data)
        try:
            # here it fails if key column does not exist in file
            data = data.drop_duplicates(combined_key_columns)
        except KeyError as e:
            pk_missing_name = str(e).split("'")[1] # cant find a nicer way of extracting missing column name
            activity.update_activity(
                ActivityStatus.FAILED,
                last_message=f"Primary Key column {pk_missing_name} does not exist in the input data. Aborting upsert.",
                progress=100,
            )
            return 0, 0

        after_rows = len(data)
        if after_rows < before_rows:
            self.log.info(
                f"Cannot upsert duplicate rows: Removed {before_rows - after_rows} duplicates from upsert input data"
            )

        # Sometimes no columns match up (including for malformed
        # xlsx files saved in 3rd party tools)
        if len(data.columns) == 0:
            self.log.warning("No columns to import")
            return 0, 0

        updated_rows = 0
        inserted_rows = 0

        # Want to either make a transaction, or a nested transaction depending on the
        # presence or absence of a user transaction (if one exists then nest another,
        # else create a root)
        with self.begin() as connection:

            # Create temporary table
            temp_table_name = "temp_table_" + str(uuid.uuid4()).replace("-", "")
            self.log.info("Moving data to temporary table: %s", temp_table_name)

            # Note: this will also clone custom columns
            create_temp_table_sql = f"""
                /* {_correlation_id} cflib.upsert */
                CREATE TEMPORARY TABLE {temp_table_name} AS
                SELECT *
                FROM {table_name}
                WITH NO DATA;
                """

            connection.execute(text(create_temp_table_sql))

            # Copy data from df to temporary table
            copy_sql = sql.SQL(
                "COPY {table} ({fields}) FROM STDIN WITH (FORMAT CSV, HEADER TRUE)"
            ).format(
                table=sql.Identifier(temp_table_name),
                fields=sql.SQL(", ").join(map(sql.Identifier, data.columns)),
            )

            # For key columns (only) convert to a placeholder value
            for column in combined_key_columns:
                if column in data.columns:
                    data[column].fillna(PLACEHOLDER_VALUE, inplace=True)

            with connection.connection.cursor() as cursor:
                start_time = time.time()
                cursor.copy_expert(copy_sql, StringIO(data.to_csv(index=False)))
                self.log.info(
                    f"Copy data to {temp_table_name} took {time.time() - start_time} seconds"
                )
                del data

                # Now upsert from temporary table to final table

                # Note: Looked at ON CONFLICT for upsert here, but not possible without
                # defining constraints on target table so doing insert and update separately

                all_columns_list = ", ".join([f'"{col_name}"' for col_name in all_column_names])

                if combined_key_columns:
                    update_column_list = ", ".join(
                        [
                            f'"{col_name}" = "{temp_table_name}"."{col_name}"'
                            for col_name in update_columns
                        ]
                    )
                    key_condition = " AND ".join(
                        [
                            f'COALESCE("{table_name}"."{key_col}", \'{PLACEHOLDER_VALUE}\') = COALESCE("{temp_table_name}"."{key_col}", \'{PLACEHOLDER_VALUE}\')'
                            for key_col in combined_key_columns
                        ]
                    )

                    cursor.execute(
                        f"SET idle_in_transaction_session_timeout = '{CFLIB_IDLE_TRANSACTION_TIMEOUT}s';"
                    )  # Idle transaction timeout in seconds

                    # Pre-locking is used here to prevent collision with other table usage or multiple upserts
                    # Will fail fast (exception) if all locks cannot be obtained
                    lock_query = f"""
                        /* {_correlation_id} cflib.upsert.lock */
                        SELECT 1
                        FROM {table_name}
                        FOR UPDATE NOWAIT;
                    """
                    start_time = time.time()
                    cursor.execute(lock_query)
                    updated_rows = cursor.rowcount
                    self.log.info(
                        f"Locking query took {time.time() - start_time} seconds for {table_name}"
                    )

                    # TODO: Indexing is based on the key columns required - this varies per upsert in some cases, due to
                    # allowing custom columns to be part of the index
                    _upsert_index_name = self._create_upsert_index_for_table(
                        table_name, cursor, combined_key_columns
                    )

                    # Update rows in the table that match the input data
                    update_query = f"""
                        /* {_correlation_id} cflib.upsert.update */
                        UPDATE {table_name}
                        SET {update_column_list}
                        FROM {temp_table_name}
                        WHERE {key_condition};
                    """

                    start_time = time.time()
                    cursor.execute(update_query)
                    updated_rows = cursor.rowcount
                    self.log.info(
                        f"Updated {updated_rows} rows in {table_name} in {time.time() - start_time} seconds"
                    )

                    # Remove rows that matched from temp table (safest approach in presence of duplicates in target)
                    delete_query = f"""
                        /* {_correlation_id} cflib.upsert.delete */
                        DELETE FROM {temp_table_name}
                        USING {table_name}
                        WHERE {key_condition}
                    """

                    start_time = time.time()
                    cursor.execute(delete_query)
                    deleted_rows = cursor.rowcount
                    self.log.info(
                        f"Deleted {deleted_rows} rows in {time.time() - start_time} seconds for temp_table"
                    )

                    temp_columns_list = ", ".join(
                        [f'"{temp_table_name}"."{col_name}"' for col_name in all_column_names]
                    )

                    insert_query = f"""
                        /* {_correlation_id} cflib.upsert.insert */
                        INSERT INTO {table_name} ({all_columns_list})
                        SELECT {temp_columns_list}
                        FROM {temp_table_name}
                    """

                    start_time = time.time()
                    cursor.execute(insert_query)
                    inserted_rows = cursor.rowcount
                    self.log.info(
                        f"Inserted {inserted_rows} into {table_name} in {time.time() - start_time} seconds"
                    )

                    # Finally remove the index created for upsert
                    cursor.execute(f"DROP INDEX IF EXISTS {_upsert_index_name};")

                # If no key columns, then just insert
                else:
                    insert_query = f"""
                        /* {_correlation_id} cflib.upsert.insert_only */
                        INSERT INTO {table_name} ({all_columns_list})
                        SELECT {all_columns_list}
                        FROM {temp_table_name}
                    """

                    updated_rows = 0
                    self.log.info(f"Running insert query for {table_name}")
                    cursor.execute(insert_query)
                    inserted_rows = cursor.rowcount

                self.log.info("Updated rows  = %s for %s", updated_rows, table_name)
                self.log.info("Inserted rows = %s for %s", inserted_rows, table_name)

        return updated_rows, inserted_rows



"""
    Library setup file
"""

from setuptools import setup

setup(
    name="cosmicfrog",
    include_package_data=True,
    version="0.3.59",
    description="Helpful utilities for working with Cosmic Frog models",
    url="https://cosmicfrog.com",
    author="Optilogic",
    packages=["cosmicfrog"],
    package_data={
        "cosmicfrog": [
            "anura27/*.json",
            "anura27/table_definitions/*.json",
        ],
    },
    license="MIT",
    install_requires=[
        "numpy>=1.26.4",
        "pandas>=2.2.1",
        "psycopg2-binary>=2.9.9",
        "sqlalchemy>=2.0.27",
        "opencensus-ext-azure>=1.1.7",
        "optilogic>=2.13.0",
        "PyJWT>=2.8.0",
        "httpx>=0.24.1",
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
)



"""
    Utility functions for Cosmic Frog
"""

import os
from pandas import DataFrame
from httpx import Client

# Note: Geocoding is via PROD service unless overidden
HYPNOTOAD_URL = (
    os.getenv("HYPNOTOAD_URL") or "https://api.cosmicfrog.optilogic.app/hypnotoad/cosmicfrog/v0.2"
)


class FrogUtils:
    """
    Container class for Cosmic Frog utilities
    """

    @staticmethod
    def geocode_table(
        model_name: str,
        table_name: str,
        app_key: str,
        geoprovider: str = "MapBox",
        geoapikey: str = None,
        ignore_low_confidence: bool = True,
    ):
        """
        Wrapper function for geocoding an Cosmic Frog table (in place)
        """
        table_name = table_name.lower().strip()

        if table_name not in ["customers", "facilities", "suppliers"]:
            raise ValueError("Table must be customers, facilities or suppliers")

        if not app_key:
            raise ValueError("Must supply a valid OL app key")

        headers = {"X-App-Key": app_key}

        params = {
            "model_name": model_name,
            "table_name": table_name,
            "geoprovider": geoprovider,
            "ignore_low_confidence": ignore_low_confidence,
        }

        if geoapikey:
            params["geoapikey"] = geoapikey

        with Client() as client:
            response = client.post(
                f"{HYPNOTOAD_URL}/geocoding/geocode/table",
                headers=headers,
                params=params,
            )
            response.raise_for_status()

    @staticmethod
    def geocode_data(data: DataFrame):
        """
        Wrapper function to geocode arbitrary location data and return result
        """

        print(data, HYPNOTOAD_URL)



"""
Module holding AnuraSchema class
"""

import os
from typing import List, Dict, Tuple
import importlib.resources
import json
from cachetools import cached, LRUCache

# Hold supported schema versions, and maps to the directory in which they are stored
# Key: schema from model db
# Value: json folder
SUPPORTED_SCHEMAS = {"anura_2_7": "anura27"}

LIBRARY_NAME = "cosmicfrog"


class AnuraSchema:
    """
    Static class for managing Anura schema versions
    """

    def __init__(self):
        raise RuntimeError("Do not instantiate directly - Static class")

    @staticmethod
    @cached(LRUCache(maxsize=len(SUPPORTED_SCHEMAS)))
    def get_anura_masterlist(anura_version: str) -> List:
        """
        Return the masterlist for the given Anura schema

        """
        json_path = f"{SUPPORTED_SCHEMAS[anura_version]}/anuraMasterTableList.json"

        with importlib.resources.as_file(
            importlib.resources.files(LIBRARY_NAME).joinpath(json_path)
        ) as file_path:
            with open(file_path, "r", encoding="utf-8") as file:
                return json.load(file)

    @staticmethod
    @cached(LRUCache(maxsize=2 * len(SUPPORTED_SCHEMAS)))
    def get_anura_table_names(anura_version: str, lower_case: bool = True) -> List[str]:
        """
        Fetch all table names
        """

        result = []

        for field in AnuraSchema.get_anura_masterlist(anura_version):
            table_name = field["Table"]
            result.append(table_name)

        if lower_case:
            return [s.lower() for s in result]

        return result

    @staticmethod
    @cached(LRUCache(maxsize=2 * len(SUPPORTED_SCHEMAS)))
    def get_anura_input_table_names(anura_version: str, lower_case: bool = True) -> List[str]:
        """
        Fetch input table names
        """

        result = []

        for field in AnuraSchema.get_anura_masterlist(anura_version):
            if field["Category"].startswith("Output") is False:
                result.append(field["Table"])

        if lower_case:
            return [s.lower() for s in result]

        return result

    @staticmethod
    @cached(LRUCache(maxsize=2 * len(SUPPORTED_SCHEMAS)))
    def get_anura_output_table_names(anura_version: str, lower_case: bool = True) -> List[str]:
        """
        Return output  table names
        """

        result = []

        for field in AnuraSchema.get_anura_masterlist(anura_version):
            if field["Category"].startswith("Output"):
                result.append(field["Table"])

        if lower_case:
            return [s.lower() for s in result]

        return result

    @staticmethod
    def get_anura_keys(anura_version: str, table_name: str) -> List[str]:
        """
        Get table keys defined in Anura schema
        """

        anura_keys, _ = AnuraSchema._get_anura_key_and_column_dicts(anura_version)

        return anura_keys[table_name]

    @staticmethod
    def get_anura_columns(anura_version: str, table_name: str) -> List[str]:
        """
        Get table coluimns defined in Anura schema
        """
        _, anura_columns = AnuraSchema._get_anura_key_and_column_dicts(anura_version)

        return anura_columns[table_name]

    @staticmethod
    @cached(LRUCache(maxsize=2))
    def _get_anura_key_and_column_dicts(
        anura_version: str,
    ) -> Tuple[Dict[str, List[str]], Dict[str, List[str]]]:
        """
        Extract keys and columns from json definitions
        """

        json_path = f"{SUPPORTED_SCHEMAS[anura_version]}/table_definitions"

        with importlib.resources.as_file(
            importlib.resources.files(LIBRARY_NAME).joinpath(json_path)
        ) as file_path:

            anura_keys = {}
            anura_cols = {}

            # Iterate over each file in the directory
            for filename in os.listdir(file_path):

                if not filename.endswith(".json"):
                    continue

                filepath = os.path.join(file_path, filename)

                with open(filepath, "r", encoding="utf-8") as f:
                    data = json.load(f)

                table_name = data.get("TableName").lower()

                # Extract the column names where "PK" is "Yes"
                anura_cols[table_name] = [
                    field["Column Name"].lower() for field in data.get("fields", [])
                ]
                anura_keys[table_name] = [
                    field["Column Name"].lower()
                    for field in data.get("fields", [])
                    if field.get("PK") == "Yes"
                ]

            return anura_keys, anura_cols



"""
    Easy logging wrapper for cosmicfrog library, supports Insights logging if available
"""

import logging
import os
import sys
from logging import Logger

app_logger = {}


def get_logger(console_only: bool = False) -> Logger:
    """
    Gets a appropriate Logger instance (configured, connected to Insights as appropriate)
    """

    current_pid = str(os.getpid())

    if current_pid in app_logger:
        return app_logger[current_pid]
    else:
        log_level = os.getenv("FROG_LOG_LEVEL") or logging.DEBUG
        log_level = int(log_level)

        logger = logging.getLogger(current_pid)
        app_logger[current_pid] = logger

        # Add log to Insights (OL internal only)
        if not console_only:
            insights_connection = os.getenv("APPLICATIONINSIGHTS_CONNECTION_STRING")

            if insights_connection:
                # Moved import here as very noisy during load (creating warnings for tests), due to deprecated features
                # in dependencies
                # TODO: Replace with opentelemetry
                # pip install opentelemetry-api
                # pip install opentelemetry-sdk
                # pip install opentelemetry-exporter-azure-monitor
                from opencensus.ext.azure.log_exporter import (
                    AzureLogHandler,
                )  # pylint: disable=import-outside-toplevel

                logger.addHandler(AzureLogHandler(connection_string=insights_connection))

        # Add log to console
        stdhandler = logging.StreamHandler(sys.stdout)
        stdhandler.setLevel(log_level)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        stdhandler.setFormatter(formatter)
        logger.addHandler(stdhandler)

        logger.setLevel(log_level)

    return logger



"""
    Helper for implementing model activity updates
"""

import os
import traceback
from typing import Optional
from logging import Logger
from enum import Enum
from httpx import AsyncClient, Client
from .sync_wrapper import sync_wrapper

CF_ACTIVITY_URL = os.getenv("CF_ACTIVITY_URL")


class ActivityStatus(Enum):
    """
    Model activity statuses.
    PENDING -> STARTED -> Terminal(COMPLETED or FAILED)
    """

    PENDING = "pending"
    STARTED = "started"
    COMPLETED = "completed"
    FAILED = "failed"


class ModelActivity:
    """
    Wraps model activity related Rest API
    """

    def __init__(
        self,
        logger: Logger,
        correlation_id: str,
        model_name: str,
        description: str,
        tags: str,
        app_key: str,
    ) -> None:
        # Fail fast if misconfigured
        assert CF_ACTIVITY_URL, "CF_ACTIVITY_URL is not configured"

        self.activity_id = None
        self.done = False
        self.logger = logger
        self.correlation_id = correlation_id
        self.model_name = model_name
        self.description = description
        self.tags = tags
        self.app_key = app_key

    def _get_api_header(self):
        return {"x-app-key": self.app_key, "correlation-id": self.correlation_id}

    async def create_activity_async(self) -> None:
        """
        Create a new activity
        """
        try:
            assert self.activity_id is None, "It is not possible to recreate an existing activity"

            self.logger.info(
                f"{self.correlation_id} Creating activity for model: {self.model_name}"
            )

            params = {
                "frogmodel_name": self.model_name,
                "description": self.description,
                "tags": self.tags,
            }

            async with AsyncClient() as client:
                response = await client.post(
                    f"{CF_ACTIVITY_URL}/activity",
                    headers=self._get_api_header(),
                    params=params,
                )

            if response.status_code != 200:
                raise ConnectionError(
                    f"{self.correlation_id} Unable to create activity: {response.status_code}"
                )

            result = response.json()

            self.activity_id = result["ActivityId"]

            self.logger.info(f"{self.correlation_id} Activity ID created: {self.activity_id}")

            return self.activity_id

        except Exception as e:
            # Ensure that an activity related failure does not stop the calling process
            trace_back = traceback.format_exc()
            self.logger.error(
                "%s Ignoring exception while creating activity: %s",
                self.correlation_id,
                e,
            )
            self.logger.debug(f"{self.correlation_id} {trace_back}")

    create_activity = sync_wrapper(create_activity_async)

    async def update_activity_async(
        self,
        activity_status: ActivityStatus,
        last_message: Optional[str] = None,
        progress: Optional[int] = None,
    ):
        """
        Update an existing activity
        """

        try:
            assert self.activity_id, "No activity_id. Check create_activity has been called"
            assert len(self.activity_id) == 36

            self.logger.info(f"activity_status: {activity_status}")
            self.logger.info(f"last_message: {last_message}")
            self.logger.info(f"progress: {progress}")

            if self.done:
                raise ValueError("Cannot update a closed activity")

            if activity_status in [ActivityStatus.COMPLETED, ActivityStatus.FAILED]:
                self.done = True

            params = {
                "frogmodel_name": self.model_name,
                "activity_id": self.activity_id,
                "activity_status": activity_status.value,
            }

            if progress:
                params["progress"] = progress

            if last_message:
                # Also, automatically log messages on the sender side also
                self.logger.info(f"{self.correlation_id} {last_message}")
                params["last_message"] = last_message

            async with AsyncClient() as client:
                response = await client.put(
                    f"{CF_ACTIVITY_URL}/activity",
                    headers=self._get_api_header(),
                    params=params,
                )

            if response.status_code != 200:
                raise ConnectionError(
                    f"{self.correlation_id} Unable to update activity: {response.status_code} {response.text}"
                )

            self.logger.info(f"{self.correlation_id} Activity ID updated: {self.activity_id}")
        except Exception as e:
            # Ensure that an activity related failure does not stop the calling process
            trace_back = traceback.format_exc()
            self.logger.error(
                "%s Ignoring exception while updating activity: %s",
                self.correlation_id,
                e,
            )
            self.logger.debug(f"{self.correlation_id} {trace_back}")

    update_activity = sync_wrapper(update_activity_async)


class AsyncFrogActivityHandler:
    """
    Async wrapper for Frog Model Activity notifications service

    Supports context manager style usage

    Can be used to create and update activities
    """

    # A new activity has:
    #
    # model_name: str =
    # Query(..., max_length=1024, description="The model the activity relates to"),
    #
    # description: str =
    # Query(..., max_length=1024, description="A short description of the activity"),
    #
    # tags: str = Query(..., max_length=1024, description="CSV tags for the activity"),

    def __init__(
        self,
        logger: Logger,
        correlation_id: str,
        model_name: str,
        description: str,
        tags: str,
        app_key: str,
    ) -> None:
        self.activity_id = None
        self.activity = ModelActivity(
            logger, correlation_id, model_name, description, tags, app_key
        )

        self.logger = logger

    async def __aenter__(self):
        try:
            activity_id = await self.activity.create_activity_async()
            self.activity_id = activity_id
        except Exception as e:
            self.logger.debug(
                f"{self.activity.correlation_id} Failed to create activity due to exception"
            )
            raise e

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            # If this activity has already been closed, then close automatically
            if not self.activity.done:
                if exc_type is None and exc_val is None and exc_tb is None:
                    await self.activity.update_activity_async(ActivityStatus.COMPLETED)
                else:
                    await self.activity.update_activity_async(ActivityStatus.FAILED)
                    raise exc_val

                self.activity.done = True

        except Exception as e:
            self.logger.debug(
                f"{self.activity.correlation_id} Failed to close activity due to exception"
            )
            self.logger.error(traceback.format_exc())
            raise e

        return self

    async def update_activity_async(
        self,
        activity_status: ActivityStatus,
        last_message: Optional[str] = None,
        progress: Optional[int] = None,
    ):
        """
        Update activity while in context
        """
        await self.activity.update_activity_async(activity_status, last_message, progress)


def activity_signal(
    logger: Logger,
    message: str,
    app_key: str,
    model_name: str = None,
    user_name: str = None,
    correlation_id: str = None,
):

    assert logger, "Must supply Logger"

    assert model_name or user_name, "Must supply either model or user name"

    assert app_key, "Must supply a valid Optilogic app_key"

    assert message, "Must supply a valid service message"

    headers = {"X-App-KEY": app_key}

    if correlation_id:
        headers["correlation-id"] = correlation_id

    params = {"message": message}

    if model_name:
        params["frogmodel_name"] = model_name

    if user_name:
        params["user_target"] = user_name

    url = f"{CF_ACTIVITY_URL}/signal"

    try:
        target = model_name if model_name else user_name
        logger.info(f"Sending activity signal: {message} to {target}")
        # TODO: Retry
        with Client() as client:
            response = client.post(url, headers=headers, params=params)
            response.raise_for_status()
            return response.status_code

    except Exception:  # pylint: disable=broad-except
        logger.exception(
            f"{correlation_id} Exception attempting to send activity signal, ignoring",
            stack_info=True,
            exc_info=True,
        )
